@@ -14,6 +14,7 @@
 #include "flang/Evaluate/type.h"
 #include "flang/Semantics/scope.h"
 #include "flang/Semantics/tools.h"
+#include <functional>
 #include <list>
 #include <map>
 #include <string>
@@ -38,7 +39,7 @@ static int FindLenParameterIndex(
 class RuntimeTableBuilder {
 public:
   RuntimeTableBuilder(SemanticsContext &, RuntimeDerivedTypeTables &);
-  void DescribeTypes(Scope &scope);
+  void DescribeTypes(Scope &scope, bool inSchemata);
 
 private:
   const Symbol *DescribeType(Scope &);
@@ -54,10 +55,13 @@ private:
   SomeExpr SaveNameAsPointerTarget(Scope &, const std::string &);
   const SymbolVector *GetTypeParameters(const Symbol &);
   evaluate::StructureConstructor DescribeComponent(const Symbol &,
-      const ObjectEntityDetails &, Scope &, const std::string &distinctName,
-      const SymbolVector *parameters);
+      const ObjectEntityDetails &, Scope &, Scope &,
+      const std::string &distinctName, const SymbolVector *parameters);
   evaluate::StructureConstructor DescribeComponent(
       const Symbol &, const ProcEntityDetails &, Scope &);
+  bool InitializeDataPointer(evaluate::StructureConstructorValues &,
+      const Symbol &symbol, const ObjectEntityDetails &object, Scope &scope,
+      Scope &dtScope, const std::string &distinctName);
   evaluate::StructureConstructor PackageIntValue(
       const SomeExpr &genre, std::int64_t = 0) const;
   SomeExpr PackageIntValueExpr(const SomeExpr &genre, std::int64_t = 0) const;
@@ -65,12 +69,12 @@ private:
   std::vector<evaluate::StructureConstructor> DescribeBindings(
       const Scope &dtScope, Scope &);
   void DescribeGeneric(
-      const GenericDetails &, std::vector<evaluate::StructureConstructor> &);
-  void DescribeSpecialProc(std::vector<evaluate::StructureConstructor> &,
+      const GenericDetails &, std::map<int, evaluate::StructureConstructor> &);
+  void DescribeSpecialProc(std::map<int, evaluate::StructureConstructor> &,
       const Symbol &specificOrBinding, bool isAssignment, bool isFinal,
       std::optional<GenericKind::DefinedIo>);
   void IncorporateDefinedIoGenericInterfaces(
-      std::vector<evaluate::StructureConstructor> &, SourceName,
+      std::map<int, evaluate::StructureConstructor> &, SourceName,
       GenericKind::DefinedIo, const Scope *);
 
   // Instantiated for ParamValue and Bound
@@ -92,23 +96,16 @@ private:
     if (auto constValue{evaluate::ToInt64(expr)}) {
       return PackageIntValue(explicitEnum_, *constValue);
     }
-    if (parameters) {
-      if (const auto *typeParam{
-              evaluate::UnwrapExpr<evaluate::TypeParamInquiry>(expr)}) {
-        if (!typeParam->base()) {
-          const Symbol &symbol{typeParam->parameter()};
-          if (const auto *tpd{symbol.detailsIf<TypeParamDetails>()}) {
-            if (tpd->attr() == common::TypeParamAttr::Len) {
-              return PackageIntValue(lenParameterEnum_,
-                  FindLenParameterIndex(*parameters, symbol));
-            }
-          }
+    if (expr) {
+      if (parameters) {
+        if (const Symbol * lenParam{evaluate::ExtractBareLenParameter(*expr)}) {
+          return PackageIntValue(
+              lenParameterEnum_, FindLenParameterIndex(*parameters, *lenParam));
         }
       }
-    }
-    if (expr) {
       context_.Say(location_,
-          "Specification expression '%s' is neither constant nor a length type parameter"_err_en_US,
+          "Specification expression '%s' is neither constant nor a length "
+          "type parameter"_err_en_US,
           expr->AsFortran());
     }
     return PackageIntValue(deferredEnum_);
@@ -128,17 +125,18 @@ private:
   SomeExpr deferredEnum_; // Value::Genre::Deferred
   SomeExpr explicitEnum_; // Value::Genre::Explicit
   SomeExpr lenParameterEnum_; // Value::Genre::LenParameter
-  SomeExpr assignmentEnum_; // SpecialBinding::Which::Assignment
+  SomeExpr scalarAssignmentEnum_; // SpecialBinding::Which::ScalarAssignment
   SomeExpr
       elementalAssignmentEnum_; // SpecialBinding::Which::ElementalAssignment
-  SomeExpr finalEnum_; // SpecialBinding::Which::Final
-  SomeExpr elementalFinalEnum_; // SpecialBinding::Which::ElementalFinal
-  SomeExpr assumedRankFinalEnum_; // SpecialBinding::Which::AssumedRankFinal
   SomeExpr readFormattedEnum_; // SpecialBinding::Which::ReadFormatted
   SomeExpr readUnformattedEnum_; // SpecialBinding::Which::ReadUnformatted
   SomeExpr writeFormattedEnum_; // SpecialBinding::Which::WriteFormatted
   SomeExpr writeUnformattedEnum_; // SpecialBinding::Which::WriteUnformatted
+  SomeExpr elementalFinalEnum_; // SpecialBinding::Which::ElementalFinal
+  SomeExpr assumedRankFinalEnum_; // SpecialBinding::Which::AssumedRankFinal
+  SomeExpr scalarFinalEnum_; // SpecialBinding::Which::ScalarFinal
   parser::CharBlock location_;
+  std::set<const Scope *> ignoreScopes_;
 };
 
 RuntimeTableBuilder::RuntimeTableBuilder(
@@ -151,26 +149,29 @@ RuntimeTableBuilder::RuntimeTableBuilder(
                                                        "deferred")},
       explicitEnum_{GetEnumValue("explicit")}, lenParameterEnum_{GetEnumValue(
                                                    "lenparameter")},
-      assignmentEnum_{GetEnumValue("assignment")},
+      scalarAssignmentEnum_{GetEnumValue("scalarassignment")},
       elementalAssignmentEnum_{GetEnumValue("elementalassignment")},
-      finalEnum_{GetEnumValue("final")}, elementalFinalEnum_{GetEnumValue(
-                                             "elementalfinal")},
-      assumedRankFinalEnum_{GetEnumValue("assumedrankfinal")},
       readFormattedEnum_{GetEnumValue("readformatted")},
       readUnformattedEnum_{GetEnumValue("readunformatted")},
       writeFormattedEnum_{GetEnumValue("writeformatted")},
-      writeUnformattedEnum_{GetEnumValue("writeunformatted")} {}
+      writeUnformattedEnum_{GetEnumValue("writeunformatted")},
+      elementalFinalEnum_{GetEnumValue("elementalfinal")},
+      assumedRankFinalEnum_{GetEnumValue("assumedrankfinal")},
+      scalarFinalEnum_{GetEnumValue("scalarfinal")} {
+  ignoreScopes_.insert(tables_.schemata);
+}
 
-void RuntimeTableBuilder::DescribeTypes(Scope &scope) {
-  if (&scope == tables_.schemata) {
-    return; // don't loop trying to describe a schema...
-  }
+void RuntimeTableBuilder::DescribeTypes(Scope &scope, bool inSchemata) {
+  inSchemata |= ignoreScopes_.find(&scope) != ignoreScopes_.end();
   if (scope.IsDerivedType()) {
-    DescribeType(scope);
-  } else {
-    for (Scope &child : scope.children()) {
-      DescribeTypes(child);
+    if (!inSchemata) { // don't loop trying to describe a schema
+      DescribeType(scope);
     }
+  } else {
+    scope.InstantiateDerivedTypes();
+  }
+  for (Scope &child : scope.children()) {
+    DescribeTypes(child, inSchemata);
   }
 }
 
@@ -263,11 +264,11 @@ static SomeExpr SaveNumericPointerTarget(
     object.set_shape(arraySpec);
     object.set_init(evaluate::AsGenericExpr(evaluate::Constant<T>{
         std::move(x), evaluate::ConstantSubscripts{elements}}));
-    const Symbol &symbol{
-        *scope
-             .try_emplace(
-                 name, Attrs{Attr::TARGET, Attr::SAVE}, std::move(object))
-             .first->second};
+    Symbol &symbol{*scope
+                        .try_emplace(name, Attrs{Attr::TARGET, Attr::SAVE},
+                            std::move(object))
+                        .first->second};
+    symbol.set(Symbol::Flag::CompilerCreated);
     return evaluate::AsGenericExpr(
         evaluate::Expr<T>{evaluate::Designator<T>{symbol}});
   }
@@ -300,11 +301,11 @@ static SomeExpr SaveDerivedPointerTarget(Scope &scope, SourceName name,
     object.set_init(
         evaluate::AsGenericExpr(evaluate::Constant<evaluate::SomeDerived>{
             derivedType, std::move(x), std::move(shape)}));
-    const Symbol &symbol{
-        *scope
-             .try_emplace(
-                 name, Attrs{Attr::TARGET, Attr::SAVE}, std::move(object))
-             .first->second};
+    Symbol &symbol{*scope
+                        .try_emplace(name, Attrs{Attr::TARGET, Attr::SAVE},
+                            std::move(object))
+                        .first->second};
+    symbol.set(Symbol::Flag::CompilerCreated);
     return evaluate::AsGenericExpr(
         evaluate::Designator<evaluate::SomeDerived>{symbol});
   }
@@ -312,20 +313,39 @@ static SomeExpr SaveDerivedPointerTarget(Scope &scope, SourceName name,
 
 static SomeExpr SaveObjectInit(
     Scope &scope, SourceName name, const ObjectEntityDetails &object) {
-  const Symbol &symbol{*scope
-                            .try_emplace(name, Attrs{Attr::TARGET, Attr::SAVE},
-                                ObjectEntityDetails{object})
-                            .first->second};
+  Symbol &symbol{*scope
+                      .try_emplace(name, Attrs{Attr::TARGET, Attr::SAVE},
+                          ObjectEntityDetails{object})
+                      .first->second};
   CHECK(symbol.get<ObjectEntityDetails>().init().has_value());
+  symbol.set(Symbol::Flag::CompilerCreated);
   return evaluate::AsGenericExpr(
       evaluate::Designator<evaluate::SomeDerived>{symbol});
 }
 
+template <int KIND> static SomeExpr IntExpr(std::int64_t n) {
+  return evaluate::AsGenericExpr(
+      evaluate::Constant<evaluate::Type<TypeCategory::Integer, KIND>>{n});
+}
+
 const Symbol *RuntimeTableBuilder::DescribeType(Scope &dtScope) {
   if (const Symbol * info{dtScope.runtimeDerivedTypeDescription()}) {
     return info;
   }
   const DerivedTypeSpec *derivedTypeSpec{dtScope.derivedTypeSpec()};
+  if (!derivedTypeSpec && !dtScope.IsParameterizedDerivedType() &&
+      dtScope.symbol()) {
+    // This derived type was declared (obviously, there's a Scope) but never
+    // used in this compilation (no instantiated DerivedTypeSpec points here).
+    // Create a DerivedTypeSpec now for it so that ComponentIterator
+    // will work. This covers the case of a derived type that's declared in
+    // a module but used only by clients and submodules, enabling the
+    // run-time "no initialization needed here" flag to work.
+    DerivedTypeSpec derived{dtScope.symbol()->name(), *dtScope.symbol()};
+    DeclTypeSpec &decl{
+        dtScope.MakeDerivedType(DeclTypeSpec::TypeDerived, std::move(derived))};
+    derivedTypeSpec = &decl.derivedTypeSpec();
+  }
   const Symbol *dtSymbol{
       derivedTypeSpec ? &derivedTypeSpec->typeSymbol() : dtScope.symbol()};
   if (!dtSymbol) {
@@ -334,7 +354,8 @@ const Symbol *RuntimeTableBuilder::DescribeType(Scope &dtScope) {
   auto locationRestorer{common::ScopedSet(location_, dtSymbol->name())};
   // Check for an existing description that can be imported from a USE'd module
   std::string typeName{dtSymbol->name().ToString()};
-  if (typeName.empty() || typeName[0] == '.') {
+  if (typeName.empty() ||
+      (typeName.front() == '.' && !context_.IsTempName(typeName))) {
     return nullptr;
   }
   std::string distinctName{typeName};
@@ -368,18 +389,6 @@ const Symbol *RuntimeTableBuilder::DescribeType(Scope &dtScope) {
     AddValue(
         dtValues, derivedTypeSchema_, "sizeinbytes"s, IntToExpr(sizeInBytes));
   }
-  const Symbol *parentDescObject{nullptr};
-  if (const Scope * parentScope{dtScope.GetDerivedTypeParent()}) {
-    parentDescObject = DescribeType(*const_cast<Scope *>(parentScope));
-  }
-  if (parentDescObject) {
-    AddValue(dtValues, derivedTypeSchema_, "parent"s,
-        evaluate::AsGenericExpr(evaluate::Expr<evaluate::SomeDerived>{
-            evaluate::Designator<evaluate::SomeDerived>{*parentDescObject}}));
-  } else {
-    AddValue(dtValues, derivedTypeSchema_, "parent"s,
-        SomeExpr{evaluate::NullPointer{}});
-  }
   bool isPDTinstantiation{derivedTypeSpec && &dtScope != dtSymbol->scope()};
   if (isPDTinstantiation) {
     // is PDT instantiation
@@ -393,9 +402,6 @@ const Symbol *RuntimeTableBuilder::DescribeType(Scope &dtScope) {
     AddValue(dtValues, derivedTypeSchema_, "uninstantiated"s,
         SomeExpr{evaluate::NullPointer{}});
   }
-
-  // TODO: compute typeHash
-
   using Int8 = evaluate::Type<TypeCategory::Integer, 8>;
   using Int1 = evaluate::Type<TypeCategory::Integer, 1>;
   std::vector<Int8::Scalar> kinds;
@@ -434,9 +440,9 @@ const Symbol *RuntimeTableBuilder::DescribeType(Scope &dtScope) {
           scope, SaveObjectName(".lpk."s + distinctName), std::move(lenKinds)));
   // Traverse the components of the derived type
   if (!isPDTdefinition) {
-    std::vector<evaluate::StructureConstructor> dataComponents;
+    std::vector<const Symbol *> dataComponentSymbols;
     std::vector<evaluate::StructureConstructor> procPtrComponents;
-    std::vector<evaluate::StructureConstructor> specials;
+    std::map<int, evaluate::StructureConstructor> specials;
     for (const auto &pair : dtScope) {
       const Symbol &symbol{*pair.second};
       auto locationRestorer{common::ScopedSet(location_, symbol.name())};
@@ -445,9 +451,8 @@ const Symbol *RuntimeTableBuilder::DescribeType(Scope &dtScope) {
               [&](const TypeParamDetails &) {
                 // already handled above in declaration order
               },
-              [&](const ObjectEntityDetails &object) {
-                dataComponents.emplace_back(DescribeComponent(
-                    symbol, object, scope, distinctName, parameters));
+              [&](const ObjectEntityDetails &) {
+                dataComponentSymbols.push_back(&symbol);
               },
               [&](const ProcEntityDetails &proc) {
                 if (IsProcedurePointer(symbol)) {
@@ -468,6 +473,18 @@ const Symbol *RuntimeTableBuilder::DescribeType(Scope &dtScope) {
           },
           symbol.details());
     }
+    // Sort the data component symbols by offset before emitting them
+    std::sort(dataComponentSymbols.begin(), dataComponentSymbols.end(),
+        [](const Symbol *x, const Symbol *y) {
+          return x->offset() < y->offset();
+        });
+    std::vector<evaluate::StructureConstructor> dataComponents;
+    for (const Symbol *symbol : dataComponentSymbols) {
+      auto locationRestorer{common::ScopedSet(location_, symbol->name())};
+      dataComponents.emplace_back(
+          DescribeComponent(*symbol, symbol->get<ObjectEntityDetails>(), scope,
+              dtScope, distinctName, parameters));
+    }
     AddValue(dtValues, derivedTypeSchema_, "component"s,
         SaveDerivedPointerTarget(scope, SaveObjectName(".c."s + distinctName),
             std::move(dataComponents),
@@ -490,12 +507,10 @@ const Symbol *RuntimeTableBuilder::DescribeType(Scope &dtScope) {
                 static_cast<evaluate::ConstantSubscript>(bindings.size())}));
     // Describe "special" bindings to defined assignments, FINAL subroutines,
     // and user-defined derived type I/O subroutines.
-    if (dtScope.symbol()) {
-      for (const auto &pair :
-          dtScope.symbol()->get<DerivedTypeDetails>().finals()) {
-        DescribeSpecialProc(specials, *pair.second, false /*!isAssignment*/,
-            true, std::nullopt);
-      }
+    const DerivedTypeDetails &dtDetails{dtSymbol->get<DerivedTypeDetails>()};
+    for (const auto &pair : dtDetails.finals()) {
+      DescribeSpecialProc(
+          specials, *pair.second, false /*!isAssignment*/, true, std::nullopt);
     }
     IncorporateDefinedIoGenericInterfaces(specials,
         SourceName{"read(formatted)", 15},
@@ -509,11 +524,39 @@ const Symbol *RuntimeTableBuilder::DescribeType(Scope &dtScope) {
     IncorporateDefinedIoGenericInterfaces(specials,
         SourceName{"write(unformatted)", 18},
         GenericKind::DefinedIo::WriteUnformatted, &scope);
+    // Pack the special procedure bindings in ascending order of their "which"
+    // code values, and compile a little-endian bit-set of those codes for
+    // use in O(1) look-up at run time.
+    std::vector<evaluate::StructureConstructor> sortedSpecials;
+    std::uint32_t specialBitSet{0};
+    for (auto &pair : specials) {
+      auto bit{std::uint32_t{1} << pair.first};
+      CHECK(!(specialBitSet & bit));
+      specialBitSet |= bit;
+      sortedSpecials.emplace_back(std::move(pair.second));
+    }
     AddValue(dtValues, derivedTypeSchema_, "special"s,
         SaveDerivedPointerTarget(scope, SaveObjectName(".s."s + distinctName),
-            std::move(specials),
+            std::move(sortedSpecials),
             evaluate::ConstantSubscripts{
                 static_cast<evaluate::ConstantSubscript>(specials.size())}));
+    AddValue(dtValues, derivedTypeSchema_, "specialbitset"s,
+        IntExpr<4>(specialBitSet));
+    // Note the presence/absence of a parent component
+    AddValue(dtValues, derivedTypeSchema_, "hasparent"s,
+        IntExpr<1>(dtScope.GetDerivedTypeParent() != nullptr));
+    // To avoid wasting run time attempting to initialize derived type
+    // instances without any initialized components, analyze the type
+    // and set a flag if there's nothing to do for it at run time.
+    AddValue(dtValues, derivedTypeSchema_, "noinitializationneeded"s,
+        IntExpr<1>(
+            derivedTypeSpec && !derivedTypeSpec->HasDefaultInitialization()));
+    // Similarly, a flag to short-circuit destruction when not needed.
+    AddValue(dtValues, derivedTypeSchema_, "nodestructionneeded"s,
+        IntExpr<1>(derivedTypeSpec && !derivedTypeSpec->HasDestruction()));
+    // Similarly, a flag to short-circuit finalization when not needed.
+    AddValue(dtValues, derivedTypeSchema_, "nofinalizationneeded"s,
+        IntExpr<1>(derivedTypeSpec && !IsFinalizable(*derivedTypeSpec)));
   }
   dtObject.get<ObjectEntityDetails>().set_init(MaybeExpr{
       StructureExpr(Structure(derivedTypeSchema_, std::move(dtValues)))});
@@ -559,11 +602,6 @@ const DeclTypeSpec &RuntimeTableBuilder::GetSchema(
   return *spec;
 }
 
-template <int KIND> static SomeExpr IntExpr(std::int64_t n) {
-  return evaluate::AsGenericExpr(
-      evaluate::Constant<evaluate::Type<TypeCategory::Integer, KIND>>{n});
-}
-
 SomeExpr RuntimeTableBuilder::GetEnumValue(const char *name) const {
   const Symbol &symbol{GetSchemaSymbol(name)};
   auto value{evaluate::ToInt64(symbol.get<ObjectEntityDetails>().init())};
@@ -579,6 +617,7 @@ Symbol &RuntimeTableBuilder::CreateObject(
       Attrs{Attr::TARGET, Attr::SAVE}, std::move(object))};
   CHECK(pair.second);
   Symbol &result{*pair.first->second};
+  result.set(Symbol::Flag::CompilerCreated);
   return result;
 }
 
@@ -589,7 +628,7 @@ SourceName RuntimeTableBuilder::SaveObjectName(const std::string &name) {
 SomeExpr RuntimeTableBuilder::SaveNameAsPointerTarget(
     Scope &scope, const std::string &name) {
   CHECK(!name.empty());
-  CHECK(name.front() != '.');
+  CHECK(name.front() != '.' || context_.IsTempName(name));
   ObjectEntityDetails object;
   auto len{static_cast<common::ConstantSubscript>(name.size())};
   if (const auto *spec{scope.FindType(DeclTypeSpec{CharacterTypeSpec{
@@ -599,24 +638,26 @@ SomeExpr RuntimeTableBuilder::SaveNameAsPointerTarget(
     object.set_type(scope.MakeCharacterType(
         ParamValue{len, common::TypeParamAttr::Len}, KindExpr{1}));
   }
-  using Ascii = evaluate::Type<TypeCategory::Character, 1>;
+  using evaluate::Ascii;
   using AsciiExpr = evaluate::Expr<Ascii>;
   object.set_init(evaluate::AsGenericExpr(AsciiExpr{name}));
-  const Symbol &symbol{
-      *scope
-           .try_emplace(SaveObjectName(".n."s + name),
-               Attrs{Attr::TARGET, Attr::SAVE}, std::move(object))
-           .first->second};
+  Symbol &symbol{*scope
+                      .try_emplace(SaveObjectName(".n."s + name),
+                          Attrs{Attr::TARGET, Attr::SAVE}, std::move(object))
+                      .first->second};
+  symbol.set(Symbol::Flag::CompilerCreated);
   return evaluate::AsGenericExpr(
       AsciiExpr{evaluate::Designator<Ascii>{symbol}});
 }
 
 evaluate::StructureConstructor RuntimeTableBuilder::DescribeComponent(
     const Symbol &symbol, const ObjectEntityDetails &object, Scope &scope,
-    const std::string &distinctName, const SymbolVector *parameters) {
+    Scope &dtScope, const std::string &distinctName,
+    const SymbolVector *parameters) {
   evaluate::StructureConstructorValues values;
+  auto &foldingContext{context_.foldingContext()};
   auto typeAndShape{evaluate::characteristics::TypeAndShape::Characterize(
-      object, context_.foldingContext())};
+      symbol, foldingContext)};
   CHECK(typeAndShape.has_value());
   auto dyType{typeAndShape->type()};
   const auto &shape{typeAndShape->shape()};
@@ -632,8 +673,21 @@ evaluate::StructureConstructor RuntimeTableBuilder::DescribeComponent(
   }
   AddValue(values, componentSchema_, "offset"s, IntExpr<8>(symbol.offset()));
   // CHARACTER length
-  const auto &len{typeAndShape->LEN()};
+  auto len{typeAndShape->LEN()};
+  if (const semantics::DerivedTypeSpec *
+      pdtInstance{dtScope.derivedTypeSpec()}) {
+    auto restorer{foldingContext.WithPDTInstance(*pdtInstance)};
+    len = Fold(foldingContext, std::move(len));
+  }
   if (dyType.category() == TypeCategory::Character && len) {
+    // Ignore IDIM(x) (represented as MAX(0, x))
+    if (const auto *clamped{evaluate::UnwrapExpr<
+            evaluate::Extremum<evaluate::SubscriptInteger>>(*len)}) {
+      if (clamped->ordering == evaluate::Ordering::Greater &&
+          clamped->left() == evaluate::Expr<evaluate::SubscriptInteger>{0}) {
+        len = clamped->right();
+      }
+    }
     AddValue(values, componentSchema_, "characterlen"s,
         evaluate::AsGenericExpr(GetValue(len, parameters)));
   } else {
@@ -687,10 +741,9 @@ evaluate::StructureConstructor RuntimeTableBuilder::DescribeComponent(
   // Shape information
   int rank{evaluate::GetRank(shape)};
   AddValue(values, componentSchema_, "rank"s, IntExpr<1>(rank));
-  if (rank > 0) {
+  if (rank > 0 && !IsAllocatable(symbol) && !IsPointer(symbol)) {
     std::vector<evaluate::StructureConstructor> bounds;
     evaluate::NamedEntity entity{symbol};
-    auto &foldingContext{context_.foldingContext()};
     for (int j{0}; j < rank; ++j) {
       bounds.emplace_back(GetValue(std::make_optional(evaluate::GetLowerBound(
                                        foldingContext, entity, j)),
@@ -713,12 +766,9 @@ evaluate::StructureConstructor RuntimeTableBuilder::DescribeComponent(
     AddValue(values, componentSchema_, "genre"s, GetEnumValue("allocatable"));
   } else if (IsPointer(symbol)) {
     AddValue(values, componentSchema_, "genre"s, GetEnumValue("pointer"));
-    hasDataInit = object.init().has_value();
-    if (hasDataInit) {
-      AddValue(values, componentSchema_, "initialization"s,
-          SomeExpr{*object.init()});
-    }
-  } else if (IsAutomaticObject(symbol)) {
+    hasDataInit = InitializeDataPointer(
+        values, symbol, object, scope, dtScope, distinctName);
+  } else if (IsAutomatic(symbol)) {
     AddValue(values, componentSchema_, "genre"s, GetEnumValue("automatic"));
   } else {
     AddValue(values, componentSchema_, "genre"s, GetEnumValue("data"));
@@ -754,6 +804,71 @@ evaluate::StructureConstructor RuntimeTableBuilder::DescribeComponent(
   return {DEREF(procPtrSchema_.AsDerived()), std::move(values)};
 }
 
+// Create a static pointer object with the same initialization
+// from whence the runtime can memcpy() the data pointer
+// component initialization.
+// Creates and interconnects the symbols, scopes, and types for
+//   TYPE :: ptrDt
+//     type, POINTER :: name
+//   END TYPE
+//   TYPE(ptrDt), TARGET, SAVE :: ptrInit = ptrDt(designator)
+// and then initializes the original component by setting
+//   initialization = ptrInit
+// which takes the address of ptrInit because the type is C_PTR.
+// This technique of wrapping the data pointer component into
+// a derived type instance disables any reason for lowering to
+// attempt to dereference the RHS of an initializer, thereby
+// allowing the runtime to actually perform the initialization
+// by means of a simple memcpy() of the wrapped descriptor in
+// ptrInit to the data pointer component being initialized.
+bool RuntimeTableBuilder::InitializeDataPointer(
+    evaluate::StructureConstructorValues &values, const Symbol &symbol,
+    const ObjectEntityDetails &object, Scope &scope, Scope &dtScope,
+    const std::string &distinctName) {
+  if (object.init().has_value()) {
+    SourceName ptrDtName{SaveObjectName(
+        ".dp."s + distinctName + "."s + symbol.name().ToString())};
+    Symbol &ptrDtSym{
+        *scope.try_emplace(ptrDtName, Attrs{}, UnknownDetails{}).first->second};
+    ptrDtSym.set(Symbol::Flag::CompilerCreated);
+    Scope &ptrDtScope{scope.MakeScope(Scope::Kind::DerivedType, &ptrDtSym)};
+    ignoreScopes_.insert(&ptrDtScope);
+    ObjectEntityDetails ptrDtObj;
+    ptrDtObj.set_type(DEREF(object.type()));
+    ptrDtObj.set_shape(object.shape());
+    Symbol &ptrDtComp{*ptrDtScope
+                           .try_emplace(symbol.name(), Attrs{Attr::POINTER},
+                               std::move(ptrDtObj))
+                           .first->second};
+    DerivedTypeDetails ptrDtDetails;
+    ptrDtDetails.add_component(ptrDtComp);
+    ptrDtSym.set_details(std::move(ptrDtDetails));
+    ptrDtSym.set_scope(&ptrDtScope);
+    DeclTypeSpec &ptrDtDeclType{
+        scope.MakeDerivedType(DeclTypeSpec::Category::TypeDerived,
+            DerivedTypeSpec{ptrDtName, ptrDtSym})};
+    DerivedTypeSpec &ptrDtDerived{DEREF(ptrDtDeclType.AsDerived())};
+    ptrDtDerived.set_scope(ptrDtScope);
+    ptrDtDerived.CookParameters(context_.foldingContext());
+    ptrDtDerived.Instantiate(scope);
+    ObjectEntityDetails ptrInitObj;
+    ptrInitObj.set_type(ptrDtDeclType);
+    evaluate::StructureConstructorValues ptrInitValues;
+    AddValue(
+        ptrInitValues, ptrDtDeclType, symbol.name().ToString(), *object.init());
+    ptrInitObj.set_init(evaluate::AsGenericExpr(
+        Structure(ptrDtDeclType, std::move(ptrInitValues))));
+    AddValue(values, componentSchema_, "initialization"s,
+        SaveObjectInit(scope,
+            SaveObjectName(
+                ".di."s + distinctName + "."s + symbol.name().ToString()),
+            ptrInitObj));
+    return true;
+  } else {
+    return false;
+  }
+}
+
 evaluate::StructureConstructor RuntimeTableBuilder::PackageIntValue(
     const SomeExpr &genre, std::int64_t n) const {
   evaluate::StructureConstructorValues xs;
@@ -813,7 +928,7 @@ RuntimeTableBuilder::DescribeBindings(const Scope &dtScope, Scope &scope) {
 }
 
 void RuntimeTableBuilder::DescribeGeneric(const GenericDetails &generic,
-    std::vector<evaluate::StructureConstructor> &specials) {
+    std::map<int, evaluate::StructureConstructor> &specials) {
   std::visit(common::visitors{
                  [&](const GenericKind::OtherKind &k) {
                    if (k == GenericKind::OtherKind::Assignment) {
@@ -842,21 +957,39 @@ void RuntimeTableBuilder::DescribeGeneric(const GenericDetails &generic,
 }
 
 void RuntimeTableBuilder::DescribeSpecialProc(
-    std::vector<evaluate::StructureConstructor> &specials,
+    std::map<int, evaluate::StructureConstructor> &specials,
     const Symbol &specificOrBinding, bool isAssignment, bool isFinal,
     std::optional<GenericKind::DefinedIo> io) {
   const auto *binding{specificOrBinding.detailsIf<ProcBindingDetails>()};
   const Symbol &specific{*(binding ? &binding->symbol() : &specificOrBinding)};
   if (auto proc{evaluate::characteristics::Procedure::Characterize(
           specific, context_.foldingContext())}) {
-    std::uint8_t rank{0};
     std::uint8_t isArgDescriptorSet{0};
     int argThatMightBeDescriptor{0};
     MaybeExpr which;
-    if (isAssignment) { // only type-bound asst's are germane to runtime
-      CHECK(binding != nullptr);
+    if (isAssignment) {
+      // Only type-bound asst's with the same type on both dummy arguments
+      // are germane to the runtime, which needs only these to implement
+      // component assignment as part of intrinsic assignment.
+      // Non-type-bound generic INTERFACEs and assignments from distinct
+      // types must not be used for component intrinsic assignment.
       CHECK(proc->dummyArguments.size() == 2);
-      which = proc->IsElemental() ? elementalAssignmentEnum_ : assignmentEnum_;
+      const auto t1{
+          DEREF(std::get_if<evaluate::characteristics::DummyDataObject>(
+                    &proc->dummyArguments[0].u))
+              .type.type()};
+      const auto t2{
+          DEREF(std::get_if<evaluate::characteristics::DummyDataObject>(
+                    &proc->dummyArguments[1].u))
+              .type.type()};
+      if (!binding || t1.category() != TypeCategory::Derived ||
+          t2.category() != TypeCategory::Derived ||
+          t1.IsUnlimitedPolymorphic() || t2.IsUnlimitedPolymorphic() ||
+          t1.GetDerivedTypeSpec() != t2.GetDerivedTypeSpec()) {
+        return;
+      }
+      which = proc->IsElemental() ? elementalAssignmentEnum_
+                                  : scalarAssignmentEnum_;
       if (binding && binding->passName() &&
           *binding->passName() == proc->dummyArguments[1].name) {
         argThatMightBeDescriptor = 1;
@@ -880,21 +1013,15 @@ void RuntimeTableBuilder::DescribeSpecialProc(
           which = assumedRankFinalEnum_;
           isArgDescriptorSet |= 1;
         } else {
-          which = finalEnum_;
-          rank = evaluate::GetRank(typeAndShape.shape());
-          if (rank > 0) {
+          which = scalarFinalEnum_;
+          if (int rank{evaluate::GetRank(typeAndShape.shape())}; rank > 0) {
             argThatMightBeDescriptor = 1;
+            which = IntExpr<1>(ToInt64(which).value() + rank);
           }
         }
       }
     } else { // user defined derived type I/O
       CHECK(proc->dummyArguments.size() >= 4);
-      bool isArg0Descriptor{
-          !proc->dummyArguments.at(0).CanBePassedViaImplicitInterface()};
-      // N.B. When the user defined I/O subroutine is a type bound procedure,
-      // its first argument is always a descriptor, otherwise, when it was an
-      // interface, it never is.
-      CHECK(!!binding == isArg0Descriptor);
       if (binding) {
         isArgDescriptorSet |= 1;
       }
@@ -919,23 +1046,26 @@ void RuntimeTableBuilder::DescribeSpecialProc(
       isArgDescriptorSet |= 1 << (argThatMightBeDescriptor - 1);
     }
     evaluate::StructureConstructorValues values;
+    auto index{evaluate::ToInt64(which)};
+    CHECK(index.has_value());
     AddValue(
         values, specialSchema_, "which"s, SomeExpr{std::move(which.value())});
-    AddValue(values, specialSchema_, "rank"s, IntExpr<1>(rank));
     AddValue(values, specialSchema_, "isargdescriptorset"s,
         IntExpr<1>(isArgDescriptorSet));
     AddValue(values, specialSchema_, "proc"s,
         SomeExpr{evaluate::ProcedureDesignator{specific}});
-    specials.emplace_back(DEREF(specialSchema_.AsDerived()), std::move(values));
+    auto pair{specials.try_emplace(
+        *index, DEREF(specialSchema_.AsDerived()), std::move(values))};
+    CHECK(pair.second); // ensure not already present
   }
 }
 
 void RuntimeTableBuilder::IncorporateDefinedIoGenericInterfaces(
-    std::vector<evaluate::StructureConstructor> &specials, SourceName name,
+    std::map<int, evaluate::StructureConstructor> &specials, SourceName name,
     GenericKind::DefinedIo definedIo, const Scope *scope) {
   for (; !scope->IsGlobal(); scope = &scope->parent()) {
     if (auto asst{scope->find(name)}; asst != scope->end()) {
-      const Symbol &generic{*asst->second};
+      const Symbol &generic{asst->second->GetUltimate()};
       const auto &genericDetails{generic.get<GenericDetails>()};
       CHECK(std::holds_alternative<GenericKind::DefinedIo>(
           genericDetails.kind().u));
@@ -950,14 +1080,11 @@ void RuntimeTableBuilder::IncorporateDefinedIoGenericInterfaces(
 
 RuntimeDerivedTypeTables BuildRuntimeDerivedTypeTables(
     SemanticsContext &context) {
-  ModFileReader reader{context};
   RuntimeDerivedTypeTables result;
-  static const char schemataName[]{"__fortran_type_info"};
-  SourceName schemataModule{schemataName, std::strlen(schemataName)};
-  result.schemata = reader.Read(schemataModule);
+  result.schemata = context.GetBuiltinModule("__fortran_type_info");
   if (result.schemata) {
     RuntimeTableBuilder builder{context, result};
-    builder.DescribeTypes(context.globalScope());
+    builder.DescribeTypes(context.globalScope(), false);
   }
   return result;
 }
